#include "fsubagent.h"
#include "portable.h"
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#include "dprint.h"

/** The type of an entry in the symbol table. */
enum fsubagentThingType { 
	/* fsubagentRow: Row of scalars */
	SA_ROW = 'r', 
	/* fsubagentArray: Table with one integer index */
	SA_ARRAY = 'a',
	/* fsubagentArray2: Table with two integer indices */
	SA_ARRAY2 = 'b' 
};

/* 
 Cheezy hardcoded limit.  We don't expect to see any 
 tables or rows with OID's longer than 7 elements.
*/
#define MAXOIDLEN 8

/**
 An entry in the table of known SNMP rows and arrays.
 This is how we map from OID's to rows or arrays.
*/
struct fsubagentThing_s {
	/** fsubagent_END-terminated oid prefix */
	int oid[MAXOIDLEN];

	/** Discriminated union */
	enum fsubagentThingType kind;
	union {
		/* Row of scalars */
		fsubagentRow *row;

		/* Table with one integer index */
		fsubagentArray *array;

		/* Table with two integer indices */
		fsubagentArray2 *array2;

	} u;
};
typedef struct fsubagentThing_s fsubagentThing;

/* 
FIXME: Cheezy hardcoded limit.  We don't expect to see 
 more than this many tables + scalar rows.
*/
#define MAXTHINGS 32

/**
 A container for SNMP rows and arrays.
 Defined only as a partial typedef in fsubagent.h.
 The full definition is private to this file.  This is the best way to
 implement private data members in C -- don't even let the world see
 the struct contents!
*/

struct fsubagent_s {
	/* The tables or rows in this container */
	int nthings;
	fsubagentThing things[MAXTHINGS];

	/* What file to save to or load from */
	char filename[PATH_MAX];

	/* Buffer used during load or save */
	char *iobuf;
	int iobuflen;
};

/**----------------------------------------------------------------------
 Create a fsubagent.
 @return initialized fsubagent * on success, NULL on failure
-----------------------------------------------------------------------*/
fsubagent *fsubagent_new()
{
	fsubagent *this = malloc(sizeof(fsubagent));
	if (!this)
		return NULL;
	this->nthings = 0;
	this->iobuflen = 64000;
	this->iobuf = malloc(this->iobuflen);
	if (!this->iobuf) {
		free(this);
		return NULL;
	}
	return this;
}

/**----------------------------------------------------------------------
 Delete a fsubagent and release any resources it has allocated.
-----------------------------------------------------------------------*/
void fsubagent_delete(fsubagent *this)
{
	int i;
	for (i=0; i<this->nthings; i++) {
		fsubagentThing *thing = &this->things[i];
		switch (thing->kind) {
		case SA_ROW:
			fsubagentRow_delete(thing->u.row);
			break;
		case SA_ARRAY:
			fsubagentArray_delete(thing->u.array);
			break;
		default:
			break;
		}
	}
	this->nthings = 0;
	if (this->iobuf)
		free(this->iobuf);
	free(this);
}

/**----------------------------------------------------------------------
 Specify the filename used for all I/O.
 @param filename - transient pointer to filename
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
int fsubagent_setFilename(fsubagent *this, const char *filename)
{
	strcpy(this->filename, filename);
	return 0;
}

#define fsubagentThing_MAGIC 0x5354 /* ST */

static int oidLen(const int *oid)
{
	int i=0;
	while (oid[i] != fsubagent_END)
		i++;
	return i;
}

static int oidCmp(const int *oida, const int *oidb)
{
	int i=0;
	while ((oida[i] != fsubagent_END) && (oida[i] == oidb[i]))
		i++;
	return oida[i] - oidb[i];
}

static void oidCpy(int *dest, const int *src)
{
	int i=0;
	while (src[i] != fsubagent_END) {
		dest[i] = src[i];
		i++;
	}
	dest[i] = fsubagent_END;
}

#ifdef USE_DPRINT
static const char *oid2a(const int *src)
{
	static char buf[1024];
	char *dest = buf;
	int i;

	*dest++ = '\'';
	i=0;
	while (src[i] != fsubagent_END) {
		dest += sprintf(dest, "%u ", src[i]);
		i++;
	}
	*dest++ = '\'';
	*dest++ = 0;
	return buf;
}
#endif

static fsubagentThing *fsubagent_findThing(fsubagent *this, const int *oid)
{
	int i;
	for (i=0; i < this->nthings; i++) {
		if (!oidCmp(this->things[i].oid, oid)) {
			DPRINT(("fsubagent_findThing(%d,%d,%d,%d...): found\n", oid[0], oid[1], oid[2], oid[3]));
			return this->things+i;
		}
	}
	DPRINT(("fsubagent_findThing(%d,%d,%d,%d...): not found\n", oid[0], oid[1], oid[2], oid[3]));
	return NULL;
}

static int fsubagent_loadThing(fsubagent *this, const char *buf, int buflen, int *bufused)
{
	/* format: 
	 * ThingRecord := magic(2) bodylen(2) ThingType(1) oidlen(1) oid(oidlen*2) ThingBody
	 * ThingBody := { RowBody | ArrayBody }
	 */
	
	int bodylen, magic;
	enum fsubagentThingType kind;
	int oidlen;
	int oid[MAXOIDLEN];
	int err;
	int i;
	fsubagentThing *thing;
	int origlen;

	origlen = buflen;

	/* Read fixed-width part of item record */
	if (buflen < 6) {
		DPRINT(("fsubagent_loadThing: fail at line %d\n", __LINE__));
		return EINVAL;
	}
	magic = BE_BYTES_TO_INT16v(buf);
	buf += 2;
	buflen -= 2;
	if (magic != fsubagentThing_MAGIC) {
		DPRINT(("fsubagent_loadThing: fail at line %d\n", __LINE__));
		return EINVAL;
	}

	bodylen = BE_BYTES_TO_INT16v(buf);
	buf += 2;
	buflen -= 2;
	kind = 0xff & *buf++;
	buflen--;
	oidlen = 0xff & *buf++;
	buflen--;

	if (bodylen < 1 || bodylen > buflen) {
		DPRINT(("fsubagent_loadThing: fail at line %d\n", __LINE__));
		return EINVAL;
	}

	if (oidlen < 1 || oidlen >= MAXOIDLEN) {
		DPRINT(("fsubagent_loadThing: fail at line %d\n", __LINE__));
		return EINVAL;
	}

	/* Read in variable-length portion */
	for (i=0; i<oidlen; i++, buf += 2, buflen -= 2)
		oid[i] = BE_BYTES_TO_INT16v(buf);
	oid[i] = fsubagent_END;

	DPRINT(("fsubagent_loadThing: loading thing with OID %s, kind %c, bodylen %d\n", 
		oid2a(oid), kind, bodylen));

	/* Locate the thing with that oid */
	thing = fsubagent_findThing(this, oid);
	if (!thing) {
		DPRINT(("fsubagent_loadThing: fail at line %d\n", __LINE__));
		return ENOENT;
	}

	switch (kind) {
	case SA_ROW:
		err = fsubagentRow_fromBuf(thing->u.row, buf, bodylen);
		break;
	case SA_ARRAY:
		err = fsubagentArray_fromBuf(thing->u.array, buf, bodylen);
		break;
	case SA_ARRAY2:
		err = fsubagentArray2_fromBuf(thing->u.array2, buf, bodylen);
		break;
	default:
		err = ENOSYS;
	}

	if (err) {
		DPRINT(("fsubagent_loadThing: fail at line %d, err %d\n", __LINE__, err));
		return err;
	}
	/*buf += bodylen;*/
	buflen -= bodylen;

	*bufused = origlen - buflen;

	DPRINT(("fsubagent_loadThing: done\n"));
	return 0;
}

static int fsubagent_saveThing(fsubagent *this, int index, char *buf, int buflen, int *bufused)
{
	int bodylen;
	int err;
	int i;
	char *origbuf;
	fsubagentThing *thing = this->things+index;
	int origlen;
	int oidlen;

	origlen = buflen;
	origbuf = buf;

	/* Write fixed-width part of item record */
	if (buflen < 6 + MAXOIDLEN*2) {
		return E2BIG;
	}
	BE_BYTES_FROM_INT16(buf, fsubagentThing_MAGIC);
	buf += 2;
	buflen -= 2;

	/* Leave the length field blank until we know what to put there */
	buf += 2;
	buflen -= 2;

	*buf++ = thing->kind;
	buflen -= 1;

	oidlen = oidLen(thing->oid);
	*buf++ = oidlen;
	buflen -= 1;

	/* write variable-length portion */
	/* NOTE: this assumes oid components are less than 65534! */
	for (i=0; i<oidlen; i++, buf += 2, buflen -= 2)
		BE_BYTES_FROM_INT16(buf, thing->oid[i]);

	switch (thing->kind) {
	case SA_ROW:
		err = fsubagentRow_toBuf(thing->u.row, buf, buflen, &bodylen);
		break;
	case SA_ARRAY:
		err = fsubagentArray_toBuf(thing->u.array, buf, buflen, &bodylen);
		break;
	case SA_ARRAY2:
		err = fsubagentArray2_toBuf(thing->u.array2, buf, buflen, &bodylen);
		break;
	default:
		err = ENOSYS;
	}

	if (err) {
		DPRINT(("fsubagent_saveThing: error: toBuf on OID %s failed, err %d!\n", 
			oid2a(thing->oid), err));
		return err;
	}
	if (bodylen == 0) {
		DPRINT(("fsubagent_saveThing: saved thing with OID %s, kind %c, bodylen %d, skipping\n", 
			oid2a(thing->oid), thing->kind, bodylen));
		/* Nothing in that object, so don't bother! */
		*bufused = 0;
		return 0;
	}
	BE_BYTES_FROM_INT16(origbuf+2, bodylen);
	/*buf += bodylen;*/
	DPRINT(("fsubagent_saveThing: saved thing with OID %s, kind %c, bodylen %d, header ", 
		oid2a(thing->oid), thing->kind, bodylen));
	fsubagent_memdump(origbuf, origlen - buflen);
	DPRINT(("\n")); 

	buflen -= bodylen;

	*bufused = origlen - buflen;

	return 0;
}

/**----------------------------------------------------------------------
 Load new current MIB contents (or changes to MIB contents) from disk.
 Must have already created all arrays and scalar rows.
 For use by snmp daemon.
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
int fsubagent_load(fsubagent *this)
{
	off_t len;
	int n;
	int fd = open(this->filename, O_RDONLY);
	char *buf;

	DPRINT(("fsubagent_load(%p): filename %s\n", this, this->filename));
	if (fd == -1) {
		DPRINT(("fsubagent_load(%p): fail, can't open file %s\n", this, this->filename));
		return errno;
	}

	/* Get file length, do sanity check */
	len = lseek(fd, 0, SEEK_END);
	lseek(fd, 0, SEEK_SET);
	if (len > 1000000) {
		close(fd);
		DPRINT(("fsubagent_load: fail at line %d\n", __LINE__));
		return E2BIG;
	}

	/* Make sure buffer's big enough */
	if (len > this->iobuflen) {
		buf = realloc(this->iobuf, len);
		if (!buf) {
			close(fd);
			DPRINT(("fsubagent_load: fail at line %d\n", __LINE__));
			return ENOMEM;
		}
		this->iobuf = buf;
		this->iobuflen = len;
	}

	/* Read in the file */
	/* FIXME: should just memory-map it */
	n = read(fd, this->iobuf, len);
	close(fd);
	if (n != len) {
		DPRINT(("fsubagent_load: fail at line %d\n", __LINE__));
		return EIO;
	}

	/* While there is more file left, load the next oid from disk,
	 * search for it in things, and load that thing from the file.
	 */
	buf = this->iobuf;
	while (len > 0) {
		int nused;
		int err = fsubagent_loadThing(this, buf, len, &nused);
		if (err) {
			DPRINT(("fsubagent_load: fail at line %d\n", __LINE__));
			return err;
		}
		DPRINT(("fsubagent_load: thing at %d took %d\n", buf-this->iobuf, nused));
		buf += nused;
		len -= nused;
	}
	DPRINT(("fsubagent_load: returning\n"));
	return 0;
}

/**----------------------------------------------------------------------
 Save the current MIB contents to a temporary file on disk, then when
 file is complete, quickly rename it to the real filename.
 (That way fsubagent_load() never sees a partial file.)
 After the first call, might only write out changes.
 For use by monitored app.
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
int fsubagent_save(fsubagent *this)
{
	int i;
	int fd;
	char tempfname[PATH_MAX];
	int err;

	sprintf(tempfname, "%s.%d.%d.tmp", this->filename, (int)getpid(), (int)pthread_self());
	fd = open(tempfname, O_RDWR|O_TRUNC|O_CREAT, 0600);
	if (fd == -1) {
		err = errno;
		DPRINT(("fsubagent_save(%p): can't create file %s, err %d\n", this, tempfname, err));
		return err;
	}
	DPRINT(("fsubagent_save(%p): saving to file %s\n", this, tempfname));

	for (i=0; i < this->nthings; i++) {
		int nused;
		int n;
		err = fsubagent_saveThing(this, i, this->iobuf, this->iobuflen, &nused);
		if (err) {
			DPRINT(("fsubagent_save(%p): saveThing %d fails, err %d\n", this, i, err));
			close(fd);
			return err;
		}
		if (!nused)
			continue;

		n = write(fd, this->iobuf, nused);
		if (n != nused) {
			close(fd);
			return EIO;
		}

	}
	close(fd);
	if (rename(tempfname, this->filename)) {
		err = errno;
		DPRINT(("fsubagent_save(%p): can't rename file %s to %s, err %d\n", this, tempfname, this->filename, err));
		return err;
	}
	DPRINT(("fsubagent_save(%p): renamed file %s to %s\n", this, tempfname, this->filename));
	return 0;
}

/**----------------------------------------------------------------------
 See comment in fsubagent.h.
-----------------------------------------------------------------------*/
fsubagentRow *fsubagent_createScalarRow(fsubagent *this, const int *oid, const char *format)
{
	fsubagentThing *thing;
	fsubagentRowConstraints *c;

	if (this->nthings >= MAXTHINGS)
		return NULL;
	thing = fsubagent_findThing(this, oid);
	if (thing)
		return NULL;
	c = fsubagentRowConstraints_new(format);
	if (!c)
		return NULL;

	thing = this->things + this->nthings++;
	thing->kind = SA_ROW;
	thing->u.row = fsubagentRow_new(c);
	oidCpy(thing->oid, oid);

	DPRINT(("fsubagent_createScalarRow(%p, %s, %s): returning %p\n", 
		this, oid2a(oid), format, thing->u.row));
	return thing->u.row;
}

/**----------------------------------------------------------------------
 See comment in fsubagent.h.
-----------------------------------------------------------------------*/
fsubagentArray *fsubagent_createArray(fsubagent *this, const int *oid, const char *format)
{
	fsubagentThing *thing;
	fsubagentRowConstraints *c;

	if (this->nthings >= MAXTHINGS)
		return NULL;
	thing = fsubagent_findThing(this, oid);
	if (thing)
		return NULL;
	c = fsubagentRowConstraints_new(format);
	if (!c)
		return NULL;

	thing = this->things + this->nthings++;
	thing->kind = SA_ARRAY;
	thing->u.array = fsubagentArray_new(c);
	oidCpy(thing->oid, oid);
	return thing->u.array;
}

/**----------------------------------------------------------------------
 Retrieve a pointer to an already created fsubagentArray.
 @param oid = transient pointer to fsubagent_END-terminated OID suffix 
 (built from symbols defined in output of mib2h.pl)
 @return array pointer on success, NULL on failure
-----------------------------------------------------------------------*/
fsubagentArray *fsubagent_getArray(fsubagent *this, const int *oid)
{
	fsubagentThing *thing;

	thing = fsubagent_findThing(this, oid);
	if (!thing)
		return NULL;
	if (thing->kind != SA_ARRAY) {
		DPRINT(("subAgent_getArray: bad type\n"));
		return NULL;
	}
	return thing->u.array;
}

/**----------------------------------------------------------------------
 See comment in fsubagent.h.
-----------------------------------------------------------------------*/
fsubagentArray2 *fsubagent_createArray2(fsubagent *this, const int *oid, const char *format)
{
	fsubagentThing *thing;
	fsubagentRowConstraints *c;

	if (this->nthings >= MAXTHINGS)
		return NULL;
	thing = fsubagent_findThing(this, oid);
	if (thing)
		return NULL;
	c = fsubagentRowConstraints_new(format);
	if (!c)
		return NULL;

	thing = this->things + this->nthings++;
	thing->kind = SA_ARRAY2;
	thing->u.array2 = fsubagentArray2_new(c);
	oidCpy(thing->oid, oid);
	return thing->u.array2;
}

/**----------------------------------------------------------------------
 Retrieve a pointer to an already created fsubagentArray2.
 @param oid = transient pointer to fsubagent_END-terminated OID suffix 
 (built from symbols defined in output of mib2h.pl)
 @return array pointer on success, NULL on failure
-----------------------------------------------------------------------*/
fsubagentArray2 *fsubagent_getArray2(fsubagent *this, const int *oid)
{
	fsubagentThing *thing;

	thing = fsubagent_findThing(this, oid);
	if (!thing)
		return NULL;
	if (thing->kind != SA_ARRAY2) {
		DPRINT(("subAgent_getArray2: bad type\n"));
		return NULL;
	}
	return thing->u.array2;
}
