#include "fsubagentArray.h"
#include "portable.h"
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

#include "dprint.h"

/**
 A 2D array of SNMP variables (i.e. an array of fsubagentRows).
 Defined only as a partial typedef in fsubagentArray.h.  
 The full definition is private to this file.  
 This is the best way to implement private data members in C -- don't
 even let the world see the struct contents!
*/

struct fsubagentArray_s {
	/** Plain C array of rows; indexed by oid - 1; allocated by malloc */
	fsubagentRow **rows;

	/** Number of elements used in rows */
	int nused;

	/** Number of elements allocated in rows */
	int nalloc;

	/** Format of rows in this array */
	const fsubagentRowConstraints *constraints;
};

/**----------------------------------------------------------------------
 Create a fsubagentArray.
 @param constraints = stable pointer to constraints on all rows
 @return initialized fsubagentArray * on success, NULL on failure
-----------------------------------------------------------------------*/
fsubagentArray *fsubagentArray_new(const fsubagentRowConstraints *c)
{
	fsubagentArray *this = malloc(sizeof(*this));
	if (!this)
		return NULL;

	this->constraints = c;

	this->rows = 0;
	this->nused = 0;
	this->nalloc = 0;

	return this;
}

/**----------------------------------------------------------------------
 Delete a fsubagentArray and release any resources it has allocated.
-----------------------------------------------------------------------*/
void fsubagentArray_delete(fsubagentArray *this)
{
	if (this->rows) {
		int i;
		for (i=0; i<this->nused; i++)
			fsubagentRow_delete(this->rows[i]);
	}
	free(this);
}

/**----------------------------------------------------------------------
 Retrieve index of highest row currently in this array.
-----------------------------------------------------------------------*/
int fsubagentArray_getMaxRow(fsubagentArray *this)
{
	return this->nused;
}

/**----------------------------------------------------------------------
 Retrieve index of highest column in this array.
-----------------------------------------------------------------------*/
int fsubagentArray_getMaxColumn(fsubagentArray *this)
{
	return this->constraints->maxoid;
}

/**----------------------------------------------------------------------
 Retrieve pointer to a row from a 2d array

 @param index = positive index (>= 1) to row of 2d array
 @return persistant row pointer on success, NULL on failure
-----------------------------------------------------------------------*/
fsubagentRow *fsubagentArray_getRow(fsubagentArray *this, int index)
{
	if (index > 0 && index <= this->nused)
		return this->rows[index-1];
	return NULL;
}

/**----------------------------------------------------------------------
 Retrieve pointer to the row containing a given identifier, if any.
 For instance, to persist the array
	partition.pname.1 = "/var"
	partition.psize.1 = 1000
	partition.pname.2 = "/usr"
	partition.psize.2 = 1000
 properly, you'd call
	usr_row = fsubagentArray_getRowByStr(array, DIMIB_pname, "/usr", 0);
 where DIMIB_pname is the OID component for 'pname' (provided by
 the mib2h.pl mib compiler) to get a pointer for the row corresponding
 to "/usr".
 @param colindex which element of the row to compare with val
 @param val the nul-terminated string to compare with
 @param prowindex where to store index of this row in the array (or NULL)
 @return persistant row pointer on success, NULL on failure
-----------------------------------------------------------------------*/
fsubagentRow *fsubagentArray_getRowByStr(fsubagentArray *this, int colindex, const char *val, int *prowindex)
{
	int i;
	if (colindex <= 0 || colindex > this->constraints->maxoid) {
		DPRINT(("fsubagentArray_getRowByStr(%p, %d, %s): colindex out of bounds\n", 
			this, colindex, val));
		return NULL;
	}
	if (this->constraints->format[colindex-1] != fsubagentItem_STRING) {
		DPRINT(("fsubagentArray_getRowByStr(%p, %d, %s): column not string\n", 
			this, colindex, val));
		return NULL;
	}
	for (i=0; i<this->nused; i++) {
		const char *rval;
		int err = fsubagentRow_getString(this->rows[i], colindex, &rval);
		if (err) {
			DPRINT(("fsubagentArray_getRowByStr(%p, %d, %s): getString returns err %d?!\n", 
				this, colindex, val, err));
			continue;
		}
		if (!strcmp(val, rval)) {
			if (prowindex) 
				*prowindex = i + 1;
			return this->rows[i];
		}
	}
	return NULL;
}

/**----------------------------------------------------------------------
 Create a new row in this Array.

 @parameter pIndex - the index of the new row is stored in *pIndex
 @return persistant row pointer on success, NULL on failure
-----------------------------------------------------------------------*/
fsubagentRow *fsubagentArray_createRow(fsubagentArray *this, int *pIndex)
{
	fsubagentRow *newrow;
	int index;

	if (this->nused == this->nalloc) {
		/* Grow the array */
		int newnalloc = (this->nalloc+1) * 2;
		fsubagentRow **newrows = realloc(this->rows, newnalloc * sizeof(*newrows));
		if (!newrows)
			return NULL;
		this->nalloc = newnalloc;
		this->rows = newrows;
	}
	newrow = fsubagentRow_new(this->constraints);
	if (!newrow)
		return NULL;
	this->nused++;
	index = this->nused;
	this->rows[index - 1] = newrow;
	if (pIndex) *pIndex = index;
	return newrow;
}

/* See comment in fsubagentArray_fromBuf for format of array body */
#define ARR_HDR_LEN 6

/**----------------------------------------------------------------------
 Save new current array contents to buffer.
 @param buf Buffer to save to.
 @param bufsize Size of buffer in bytes.
 @param bufused Pointer to where to store number of buffer bytes used.
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
int fsubagentArray_toBuf(fsubagentArray *this, char *buf, int bufsize, int *bufused)
{
	int i;
	int retval = 0;

	*bufused = 0;

	DPRINT(("fsubagentArray_toBuf: nused %d\n", this->nused));

	for (i=0; i<this->nused; i++) {
		int len;
		int err;

		DPRINT(("fsubagentArray_toBuf: at top: row %d/%d, byte %d/%d\n", 
			i, this->nused, *bufused, bufsize));

		/* Check to make sure we're not about to overflow the output buffer */
		if (*bufused + ARR_HDR_LEN >= bufsize) {
			DPRINT(("fsubagentArray_toBuf: buffer overflow\n"));
			retval = E2BIG;
			break;
		}

		/* Save the fixed-width portion */
		BE_BYTES_FROM_INT16(buf, fsubagentArray_MAGIC);
		/* Leave the length field blank until we know what to put there */
		BE_BYTES_FROM_INT16(buf+4, i + 1);

		/* Save the variable-length portion */
		err = fsubagentRow_toBuf(this->rows[i], buf+ARR_HDR_LEN, bufsize-ARR_HDR_LEN - *bufused, &len);
		if (err) {
			DPRINT(("fsubagentArray_toBuf: toBuf err %d\n", err));
			retval = err;
			break;
		}
		/* Finally write the length field */
		BE_BYTES_FROM_INT16(buf+2, len);
		/* and skip to next record */
		buf += len + ARR_HDR_LEN;
		*bufused += len + ARR_HDR_LEN;

		DPRINT(("fsubagentArray_toBuf: at bottom: row %d/%d, byte %d/%d\n", 
			i, this->nused, *bufused, bufsize));
	}

	return retval;
}

/**----------------------------------------------------------------------
 Load new current array contents from buffer.
 Must obey constraints this array was created with.
 @param buf Buffer to load from.
 @param buflen Number of bytes of data in buffer.
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
int fsubagentArray_fromBuf(fsubagentArray *this, const char *buf, int buflen)
{
	int retval = 0;

	/* format: 
	 * ArrayBody := RowRecord+
	 * RowRecord := magic(2) len(2) index(2) RowBody
	 * Note: index is 1 based
	 */

	while (buflen > 0) {
		int magic, len, index;
		int err;
		fsubagentRow *row;

		DPRINT(("fsubagentArray_fromBuf: buflen at top: %d\n", buflen));

		/* Read fixed-width part of item record */
		if (buflen < 6) {
			retval = EINVAL;
			DPRINT(("fromBuf failed, line %d\n", __LINE__));
			break;
		}
		magic = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 2;
		if (magic != fsubagentArray_MAGIC) {
			DPRINT(("fromBuf failed, line %d\n", __LINE__));
			return EINVAL;
		}
		len = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 2;
		index = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 2;

		if (len < 1 || len > buflen) {
			retval = EINVAL;
			DPRINT(("fromBuf failed, line %d\n", __LINE__));
			break;
		}

		if (index < 1) {
			retval = EINVAL;
			DPRINT(("fromBuf failed, line %d\n", __LINE__));
			break;
		}

		row = fsubagentArray_getRow(this, index);
		if (!row) {
			// FIXME: this is a bit fragile
			int rownum;
			row = fsubagentArray_createRow(this, &rownum);
			assert(index == rownum);
		}
		if (!row) {
			retval = ENOMEM;
			DPRINT(("fromBuf failed, line %d\n", __LINE__));
			break;
		}

		/* Read in variable-length portion */
		err = fsubagentRow_fromBuf(row, buf, len);
		if (err) {
			retval = err;
			DPRINT(("fromBuf failed, line %d\n", __LINE__));
			break;
		}
		buf += len;
		buflen -= len;
		DPRINT(("fsubagentArray_fromBuf: buflen at bottom: %d\n", buflen));
	}

	return retval;
}
