#include "fsubagentArray2.h"
#include "portable.h"
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>

#include "dprint.h"

/**
 A 3D array of SNMP variables (i.e. an array of fsubagentArrays).
 Defined only as a partial typedef in fsubagentArray2.h.  
 The full definition is private to this file.  
 This is the best way to implement private data members in C -- don't
 even let the world see the struct contents!
*/

struct fsubagentArray2_s {
	/** Plain C array of Array pointers; indexed by oid - 1; allocated by malloc */
	fsubagentArray **arrays;

	/** Number of elements used in arrays */
	int nused;

	/** Number of elements allocated in arrays */
	int nalloc;

	/** Format of arrays in this array */
	const fsubagentRowConstraints *constraints;
};

/**----------------------------------------------------------------------
 Create a fsubagentArray2.
 @param constraints = stable pointer to constraints on all arrays
 @return initialized fsubagentArray2 * on success, NULL on failure
-----------------------------------------------------------------------*/
fsubagentArray2 *fsubagentArray2_new(const fsubagentRowConstraints *c)
{
	fsubagentArray2 *this = malloc(sizeof(*this));
	if (!this)
		return NULL;

	this->constraints = c;

	this->arrays = 0;
	this->nused = 0;
	this->nalloc = 0;

	return this;
}

/**----------------------------------------------------------------------
 Delete a fsubagentArray2 and release any resources it has allocated.
-----------------------------------------------------------------------*/
void fsubagentArray2_delete(fsubagentArray2 *this)
{
	if (this->arrays) {
		int i;
		for (i=0; i<this->nused; i++)
			fsubagentArray_delete(this->arrays[i]);
	}
	free(this);
}

/**----------------------------------------------------------------------
 Retrieve index of highest array currently in this array.
-----------------------------------------------------------------------*/
int fsubagentArray2_getMaxArray(fsubagentArray2 *this)
{
	return this->nused;
}

/**----------------------------------------------------------------------
 Retrieve a pointer to an already created fsubagentArray inside this
 fsubagentArray2.
 @param index = positive index (>= 1) to subarray
 @return array pointer on success, NULL on failure
-----------------------------------------------------------------------*/
fsubagentArray *fsubagentArray2_getArray(fsubagentArray2 *this, int index)
{
	if (index < 1 || index > this->nused) {
		/*DPRINT(("fsubagentArray2_getArray(%p, %d): index out of range %d..%d\n", 
			this, index, 1, this->nused));*/
		return NULL;
	}
	return this->arrays[index-1];
}

/**----------------------------------------------------------------------
 Create a subarray inside this fsubagentArray2.
 @parameter pIndex - the index of the new array is stored in *pIndex
 @return array pointer on success, NULL on failure
-----------------------------------------------------------------------*/
fsubagentArray *fsubagentArray2_createArray(fsubagentArray2 *this, int *pIndex)
{
	fsubagentArray *newarray;
	int index;

	if (this->nused == this->nalloc) {
		/* Grow the array */
		int newnalloc = (this->nalloc+1) * 2;
		fsubagentArray **newarrays = realloc(this->arrays, newnalloc * sizeof(*newarrays));
		if (!newarrays)
			return NULL;
		this->nalloc = newnalloc;
		this->arrays = newarrays;
	}
	newarray = fsubagentArray_new(this->constraints);
	if (!newarray)
		return NULL;
	this->nused++;
	index = this->nused;
	this->arrays[index - 1] = newarray;
	DPRINT(("fsubagentArray2_createArray(%p): created array at index %d, returning %p\n", 
		this, index, newarray));
	if (pIndex) *pIndex = index;
	return newarray;
}

/* See comment in fsubagentArray2_fromBuf for format of array2 body */
#define ARR2_HDR_LEN 6

/**----------------------------------------------------------------------
 Save new current array contents to buffer.
 @param buf Buffer to save to.
 @param bufsize Size of buffer in bytes.
 @param bufused Pointer to where to store number of buffer bytes used.
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
int fsubagentArray2_toBuf(fsubagentArray2 *this, char *buf, int bufsize, int *bufused)
{
	int i;
	int retval = 0;

	*bufused = 0;

	for (i=0; i<this->nused; i++) {
		int len;
		int err;

		DPRINT(("fsubagentArray2_toBuf: at top: thing %d/%d, byte %d/%d\n", 
			i, this->nused, *bufused, bufsize));

		/* Check to make sure we're not about to overflow the output buffer */
		if (*bufused + ARR2_HDR_LEN >= bufsize) {
			DPRINT(("fsubagentArray2_toBuf: buffer overflow at line %d\n", __LINE__));
			retval = E2BIG;
			break;
		}

		/* Save the fixed-width portion */
		BE_BYTES_FROM_INT16(buf, fsubagentArray2_MAGIC);
		/* Leave the length field blank until we know what to put there */
		BE_BYTES_FROM_INT16(buf + 4, i + 1);

		/* Save the variable-length portion */
		err = fsubagentArray_toBuf(this->arrays[i], buf+ARR2_HDR_LEN, bufsize-ARR2_HDR_LEN - *bufused, &len);
		if (err) {
			retval = err;
			DPRINT(("fsubagentArray2_toBuf: array_toBuf err %d at line %d\n", err, __LINE__));
			break;
		}
#if 0	/* can't do this until we do the FIXME in fromBuf */
		/* If there was nothing there, skip this record */
		if (len == 0) {
			DPRINT(("fsubagentArray2_toBuf: array_toBuf len 0, skipping\n"));
			continue;
		}
#endif
		/* Finally write the length field */
		BE_BYTES_FROM_INT16(buf + 2, len);
		DPRINT(("fsubagentArray2_toBuf: header: "));
		fsubagent_memdump(buf, 6);
		DPRINT(("\n"));
		/* and skip to next record */
		buf += len + ARR2_HDR_LEN;
		*bufused += len + ARR2_HDR_LEN;

		DPRINT(("fsubagentArray2_toBuf: at bottom: thing %d/%d, byte %d/%d\n", 
			i, this->nused, *bufused, bufsize));
	}

	return retval;
}

/**----------------------------------------------------------------------
 Load new current array contents from buffer.
 Must obey constraints this array was created with.
 @param buf Buffer to load from.
 @param buflen Number of bytes of data in buffer.
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
int fsubagentArray2_fromBuf(fsubagentArray2 *this, const char *buf, int buflen)
{
	int retval = 0;

	/* format: 
	 * ArrayBody := RowRecord+
	 * RowRecord := magic(2) len(2) index(2) RowBody
	 * Note: index is 1 based
	 */

	while (buflen > 0) {
		int magic, len, index;
		int err;
		fsubagentArray *array;

		DPRINT(("fsubagentArray2_fromBuf: at top: buflen %d\n", buflen));

		/* Read fixed-width part of item record */
		if (buflen < 6) {
			retval = EINVAL;
			DPRINT(("fromBuf failed, file %s, line %d\n", __FILE__, __LINE__));
			break;
		}
		magic = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 2;
		if (magic != fsubagentArray2_MAGIC) {
			DPRINT(("fromBuf failed, file %s, line %d\n", __FILE__, __LINE__));
			return EINVAL;
		}
		len = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 2;
		index = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 2;

		DPRINT(("fsubagentArray2_fromBuf: got len %d, index %d; buflen %d\n", len, index, buflen));

		/*if (len < 1 || len > buflen) */
		if (len < 0 || len > buflen) {
			retval = EINVAL;
			DPRINT(("fromBuf failed; len %d not in range 1..%d, file %s, line %d\n", 
				len, buflen, __FILE__, __LINE__));
			break;
		}

		if (index < 1) {
			retval = EINVAL;
			DPRINT(("fromBuf failed, file %s, line %d\n", __FILE__, __LINE__));
			break;
		}

		DPRINT(("fsubagentArray2_fromBuf: looking up array at index %d\n", index));
		array = fsubagentArray2_getArray(this, index);
		if (!array) {
			// FIXME: this is a bit fragile
			int arrnum;
			DPRINT(("fsubagentArray2_fromBuf: creating array at index %d\n", index));
			array = fsubagentArray2_createArray(this, &arrnum);
			assert(index == arrnum);
		}
		if (!array) {
			retval = ENOMEM;
			DPRINT(("fromBuf failed, file %s, line %d\n", __FILE__, __LINE__));
			break;
		}

		// We can get rid of this 'if' once we make toBuf not output zero-length records!
		if (len == 0)
			continue;

		/* Read in variable-length portion */
		err = fsubagentArray_fromBuf(array, buf, len);
		if (err) {
			retval = err;
			DPRINT(("fromBuf failed, file %s, line %d\n", __FILE__, __LINE__));
			break;
		}
		buf += len;
		buflen -= len;

		DPRINT(("fsubagentArray2_fromBuf: at bottom: buflen %d\n", buflen));
	}

	return retval;
}
