#ifndef fsubagentArray2_H
#define fsubagentArray2_H

#include "fsubagentArray.h"

#include <stdlib.h>

#ifdef __cplusplus
extern "C" {
#endif

/* First two bytes (in big endian order) written by fsubagentArray2_toBuf() */
#define fsubagentArray2_MAGIC 0x5342 /* SB */

/**----------------------------------------------------------------------
 An array of fsubagentArray; this is used to implement SNMP tables with two 
 integer indices.   The programmer uses the first index as an index 
 into fsubagentArray2 to retrieve a fsubagentArray pointer arr,
 then uses the second index as an index into arr to retrieve a row pointer;
 and finally uses the last part of the OID before the indices as an index 
 into the row.

 In the SNMP variable
    barTable.barEntry.barMember.index1.index2
 the next-to-last index (index1) is what you pass to
 fsubagentArray2_getArray() to locate the appropriate fsubagentArray.

 For instance, if one wants a table with two indices to represent the
 partition tables of a bunch of disks, e.g.

	partEntry OBJECT-TYPE
			SYNTAX PartEntry
			ACCESS read-only
			STATUS mandatory
			DESCRIPTION
				"Describes a partition table entry on a given disk"
			INDEX    { diskIndex, partIndex }
			::= { partTable 1 }

 and one had two disks, one having two partitions and one having one
 parition, the populated table might look like this in snmpwalk:
     partTable.partEntry.pStart.1.1 = 0		# first disk, first partition
     partTable.partEntry.pSize.1.1  = 1000
     partTable.partEntry.pStart.1.2 = 1000	# first disk, second partition
     partTable.partEntry.pSize.1.2  = 1000
     partTable.partEntry.pStart.2.1 = 0		# second disk, first partition
     partTable.partEntry.pSize.2.1  = 2000

 The code to implement the above might look like this:
 1. Create the overall table:
 	ptable = fsubagent_createArray2(DIMIB_partEntry_FORMAT);
	(This does fsubagentArray2_new(), then hooks the newly created table 
	up so fsubagent_save() will save it to disk.)
 2. Create a subarray for each disk:
 	disk1tab = fsubagentArray2_createArray(ptable, NULL);
 	disk2tab = fsubagentArray2_createArray(ptable, NULL);
 3. Create a row for each partition:
    disk1part1 = fsubagentArray_createRow(disk1tab);
    disk1part2 = fsubagentArray_createRow(disk1tab);
    disk2part1 = fsubagentArray_createRow(disk2tab);
 4. Populate with values:
 	fsubagentRow_setInt(disk1part1, DIMIB_pStart, 0);
 	fsubagentRow_setInt(disk1part1, DIMIB_pSize, 1000);
 	fsubagentRow_setInt(disk1part2, DIMIB_pStart, 1000);
 	fsubagentRow_setInt(disk1part2, DIMIB_pSize, 1000);
 	fsubagentRow_setInt(disk2part1, DIMIB_pStart, 0);
 	fsubagentRow_setInt(disk2part1, DIMIB_pSize, 2000);

 The only methods for use by the app programmer are getMaxArray(),
 getArray(), and createArray().  The rest of its methods are not meant to
 be called directly by the programmer; see fsubagent.h.

 The module is implemented in C using an object-oriented idiom which
 provides good encapsualation.  Since all data members of fsubagentArray2
 are private, the structure is only named here; the full definition is
 private to fsubagentArray2.c.
-----------------------------------------------------------------------*/
typedef struct fsubagentArray2_s fsubagentArray2;

/* class fsubagentArray2 { */

	/**----------------------------------------------------------------------
	 Create a fsubagentArray2.
	 @param constraints = stable pointer to constraints on all rows
	 @return initialized fsubagentArray2 * on success, NULL on failure
	-----------------------------------------------------------------------*/
	fsubagentArray2 *fsubagentArray2_new(const fsubagentRowConstraints *c);

	/**----------------------------------------------------------------------
	 Delete a fsubagentArray2 and release any resources it has allocated.
	-----------------------------------------------------------------------*/
	void fsubagentArray2_delete(fsubagentArray2 *);

	/**----------------------------------------------------------------------
	 Retrieve index of highest array currently in this array.
	-----------------------------------------------------------------------*/
	int fsubagentArray2_getMaxArray(fsubagentArray2 *);

	/**----------------------------------------------------------------------
	 Create a subarray inside this fsubagentArray2.
	 @parameter pIndex - the index of the new array is stored in *pIndex
	 @return array pointer on success, NULL on failure
	-----------------------------------------------------------------------*/
	fsubagentArray *fsubagentArray2_createArray(fsubagentArray2 *, int *pIndex);

	/**----------------------------------------------------------------------
	 Retrieve a pointer to an already created fsubagentArray inside this
	 fsubagentArray2.
	 @param index = positive index (>= 1) to subarray
	 @return array pointer on success, NULL on failure
	-----------------------------------------------------------------------*/
	fsubagentArray *fsubagentArray2_getArray(fsubagentArray2 *, int index);

	/**----------------------------------------------------------------------
	 Load new fsubagentArray2 element contents from buffer.
	 Must obey constraints this array was created with.
	 @param buf Buffer to load from.
	 @param buflen Number of bytes of data in buffer.
	 @return 0 on success, Unix error code on failure
	-----------------------------------------------------------------------*/
	int fsubagentArray2_fromBuf(fsubagentArray2 *, const char *buf, int buflen);

	/**----------------------------------------------------------------------
	 Save new fsubagentArray2 element contents to buffer.
	 @param buf Buffer to save to.
	 @param bufsize Size of buffer in bytes.
	 @param bufused Pointer to where to store number of buffer bytes used.
	 @return 0 on success, Unix error code on failure
	-----------------------------------------------------------------------*/
	int fsubagentArray2_toBuf(fsubagentArray2 *, char *buf, int bufsize, int *bufused);

/* } */

#ifdef __cplusplus
}
#endif

#endif
