#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fsubagentArray2.h"

#define CHECK(a, b) assert(a == b)
#define CHECKNE(a, b) assert(a != b)
#define STRCHECK(a, b) assert(!strcmp(a, b))

/* Length of side of square array of rows for testing */
#define SIZE 2

static void createRows(fsubagentArray2 *root)
{
	/* These indices are indexed by oid component, which is >= 1; 0th index not used */
	/* We don't really need these arrays, but they make things a bit clearer */
	fsubagentArray *arr[SIZE+1], *arrx;
	fsubagentRow *row[SIZE+1][SIZE+1];
	int i, j;

	/* Populate row pointers */
	for (i=1; i <= SIZE; i++) {
		int arrnum;
		arr[i] = fsubagentArray2_createArray(root, &arrnum);
		CHECKNE(arr[i], 0);
		CHECK(i, arrnum);
		arrx = fsubagentArray2_getArray(root, i);
		CHECK(arr[i], arrx);
		for (j=1; j <= SIZE; j++) {
			int rownum;
			row[i][j] = fsubagentArray_createRow(arr[i], &rownum);
			CHECK(j, rownum);
			CHECKNE(row[i][j], 0);
		}
	}
}

/** 
 Store a test pattern in the Array2 in the rows created by createRows().
 For convenience, use the oid as the value, kinda 
*/
static void storeTestPattern(fsubagentArray2 *root)
{
	/* These indices are indexed by oid component, which is >= 1; 0th index not used */
	/* We don't really need these arrays, but they make things a bit clearer */
	fsubagentArray *arr[SIZE+1];
	fsubagentRow *row[SIZE+1][SIZE+1];
	int i, j;

	/* Populate row pointers */
	for (i=1; i <= SIZE; i++) {
		arr[i] = fsubagentArray2_getArray(root, i);
		CHECKNE(arr[i], 0);
		for (j=1; j <= SIZE; j++) {
			row[i][j] = fsubagentArray_getRow(arr[i], j);
			CHECKNE(row[i][j], 0);
		}
	}

	/* Store test pattern */
	for (i=1; i <= SIZE; i++) {
		for (j=1; j <= SIZE; j++) {
			int k;
			char sval[256];
			/* We happen to know that format[0] and format[1] are 'i' */
			for (k=1; k <= 2; k++) {
				int ival = (i << 8) + (j << 4) + k;
				CHECK(0, fsubagentRow_setInt(row[i][j], k, ival));
			}
			/* We happen to know that format[2] is 's' */
			for (k=3; k <= 3; k++) {
				sprintf(sval, "%1d%1d%1d", i, j, k);
				CHECK(0, fsubagentRow_setString(row[i][j], k, sval));
			}
		}
	}
}

/** 
 Make sure the test pattern set by storeTestPattern() is still there
 in the Array2.
*/
static void checkTestPattern(fsubagentArray2 *root)
{
	/* These indices are indexed by oid component, which is >= 1; 0th index not used */
	/* We don't really need these arrays, but they make things a bit clearer */
	fsubagentArray *arr[SIZE+1];
	fsubagentRow *row[SIZE+1][SIZE+1];
	int i, j;

	/* Populate row pointers */
	for (i=1; i <= SIZE; i++) {
		arr[i] = fsubagentArray2_getArray(root, i);
		CHECKNE(arr[i], 0);
		for (j=1; j <= SIZE; j++) {
			row[i][j] = fsubagentArray_getRow(arr[i], j);
		}
	}

	/* Check test pattern */
	for (i=1; i <= SIZE; i++) {
		for (j=1; j <= SIZE; j++) {
			int k;
			for (k=1; k<=2; k++) {
				int ival;
				int want = (i << 8) + (j << 4) + k;
				CHECK(0, fsubagentRow_getInt(row[i][j], k, &ival));
				CHECK(want, ival);
			}
			for (k=3; k<=3; k++) {
				const char *sval;
				char swant[256];
				sprintf(swant, "%1d%1d%1d", i, j, k);
				CHECK(0, fsubagentRow_getString(row[i][j], k, &sval));
				CHECK(0, strcmp(sval, swant));
				/* Make sure associative lookup works */
				CHECK(row[i][j], fsubagentArray_getRowByStr(arr[i], k, swant, 0));
			}
		}
	}

	/* Make sure we don't find things that aren't there */
	CHECK(NULL,  fsubagentArray_getRowByStr(arr[1], 3, "blort", 0));
}

static void saveSomeData(void)
{
	char buf[64000];
	int buflen;
	int i;
	fsubagentRowConstraints constraints;
	fsubagentArray2 *root;

	constraints.format = "iis";
	constraints.maxoid = strlen(constraints.format);

	root = fsubagentArray2_new(&constraints);
	CHECKNE(0, root);

	createRows(root);
	storeTestPattern(root);
	checkTestPattern(root);

	/* Save to buffer */
	buflen = 0;
	CHECK(0, fsubagentArray2_toBuf(root, buf, sizeof(buf), &buflen));

	fsubagentArray2_delete(root);

	printf("Info: buf = ");
	for (i=0; i<buflen; i++)
		printf("%02x ", 0xff & buf[i]);
	printf("\n");

	/* Load from buffer.  No need to create rows first, fromBuf will do it. */
	root = fsubagentArray2_new(&constraints);
	CHECKNE(0, root);
	CHECK(0, fsubagentArray2_fromBuf(root, buf, buflen));

	/* Retrieve values and make sure they're what we set earlier */
	checkTestPattern(root);

	fsubagentArray2_delete(root);
}

int main(int argc, char **argv)
{
	(void) argc;
	(void) argv;
	saveSomeData();
	printf("fsubagentArray2_test: pass\n");
	return 0;
}
