#include "fsubagentRow.h"
#include "portable.h"
#include <assert.h>
#include <errno.h>
#include <pthread.h>
#include <string.h>
#include <stdio.h>

#include "dprint.h"

/**
 A single SNMP variable.
 Can't interpret without the matching fsubagentItemType,
 which is stored in fsubagentRow.constraints->format[].
*/
struct fsubagentItem_s {
	/* stored in fsubagentRow.constraints->format[], not here,
	 * since many rows can share same format 
	 */
	/* enum fsubagentItemType kind; */

	union {
		int intval;
		struct {
			int len;		/* number of bytes allocated for buf */
			char *buf;		/* allocated with malloc */
		} stringval;
	} u;
};
typedef struct fsubagentItem_s fsubagentItem;

/**
 A 1D array of SNMP variables.
 Defined only as a partial typedef in fsubagentRow.h.  
 The full definition is private to this file.  
 This is the best way to implement private data members in C -- don't
 even let the world see the struct contents!
*/

struct fsubagentRow_s {
	/** mutex to protect contents against thread conflicts */
	pthread_mutex_t mutex;   

	/** Where variables are stored; indexed by oid - 1; allocated by malloc */
	fsubagentItem **vars;

	/** Number of elements allocated in vars */
	int nvars;

	/** Format for the row */
	const fsubagentRowConstraints *constraints;
};

/** Given a transient pointer to the row's format, create a constraints object. */
fsubagentRowConstraints *fsubagentRowConstraints_new(const char *format)
{
	fsubagentRowConstraints *this;
	this = malloc(sizeof(*this));
	if (!this)
		return NULL;
	this->format = strdup(format);
	if (!this->format) {
		free(this);
		return NULL;
	}
	this->maxoid = strlen(format);
	return this;
}

/**----------------------------------------------------------------------
 Create a fsubagentItem containing no value.
-----------------------------------------------------------------------*/
static fsubagentItem *fsubagentItem_new(enum fsubagentItemType kind)
{
	/* FIXME: should decide how much to malloc based on kind */
	fsubagentItem *this = malloc(sizeof(*this));
	if (!this) {
		DPRINT(("fsubagentItem_new: fail\n"));
		return NULL;
	}
	switch (kind) {
	case fsubagentItem_INT: 
		this->u.intval = 0;
		break;
	case fsubagentItem_STRING: 
		this->u.stringval.len = 1;
		this->u.stringval.buf = malloc(1);
		this->u.stringval.buf[0] = 0;
		break;
	default:
		free(this);
		DPRINT(("fsubagentItem_new: bad kind\n"));
		return NULL;
	}
	DPRINT(("fsubagentItem_new: returning %p\n", this));
	return this;
}

/**----------------------------------------------------------------------
 Delete a fsubagentItem and release any resources it has allocated.
 @param kind corresponding element of fsubagentRow.constraints->format[]
-----------------------------------------------------------------------*/
static void fsubagentItem_delete(fsubagentItem *this, enum fsubagentItemType kind)
{
	DPRINT(("fsubagentItem_new(%p): entry\n", this));
	if (kind == fsubagentItem_STRING) {
		if (this->u.stringval.buf)
			free(this->u.stringval.buf);
		this->u.stringval.buf = NULL;
	}
	free(this);
	DPRINT(("fsubagentItem_delete(%p): done\n", this));
}

/**----------------------------------------------------------------------
 Copy the given string into the given fsubagentItem (which must already
 have type STRING).
 len does not include any nul terminating byte.
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
static int fsubagentItem_copyString(fsubagentItem *this, const char *buf, int len)
{
	DPRINT(("fsubagentItem_copyString(%p, %s, %d): entry\n", this, buf, len));
	if (this->u.stringval.len <= len) {
		char *tmp = realloc(this->u.stringval.buf, len + 1);
		if (!tmp)
			return ENOMEM;
		this->u.stringval.buf = tmp;
	}
	this->u.stringval.len = len + 1;
	memcpy(this->u.stringval.buf, buf, len);
	this->u.stringval.buf[len] = 0;
	DPRINT(("fsubagentItem_copyString(%p, %s, %d): success\n", this, buf, len));
	return 0;
}

/**----------------------------------------------------------------------
 Load new current variable contents from buffer.
 For use by fsubagentRow_fromBuf only.
 @param kind corresponding element of fsubagentRow.constraints->format[]
 @param buf Buffer to load from.
 @param buflen Number of bytes of data in buffer.  Must be 4 for int type.
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
static int fsubagentItem_fromBuf(fsubagentItem *this, enum fsubagentItemType kind, const char *buf, int buflen)
{
	/* format: value(buflen)
	 * Type is implicit; it's stored in fsubagentItemRow.constraints->format[]
	 * If type is fsubagentItem_STRING, buf and buflen do NOT include 
	 * nul terminating byte; terminating nul is appended here.
	 * Byte order is big endian.
	 */

	switch (kind) {
	case fsubagentItem_INT:
		if (buflen != sizeof(int)) {
			/* bad format */
			return EINVAL;
		}
		this->u.intval = BE_BYTES_TO_INT32v(buf);
		break;
	case fsubagentItem_STRING:
		return fsubagentItem_copyString(this, buf, buflen);
	default:
		return EINVAL;
	}

	return 0;
}

/**----------------------------------------------------------------------
 Save current variable contents to buffer.
 For use by fsubagentRow_toBuf only.
 @param kind corresponding element of fsubagentRow.constraints->format[]
 @param buf Buffer to save to.
 @param buflen Number of bytes of data in buffer.
 @return ***negative Unix error code on failure, number of bytes consumed on success***
-----------------------------------------------------------------------*/
static int fsubagentItem_toBuf(fsubagentItem *this, enum fsubagentItemType kind, char *buf, int buflen)
{
	/* See comment in fsubagentItem_fromBuf for format */

	if (kind == fsubagentItem_UNDEF) {
		 /* no point in saving empty variables */
		return 0;
	}

	switch (kind) {
	case fsubagentItem_INT:
		if (buflen < 1 + (int)sizeof(int)) {
			return -E2BIG;
		}
		BE_BYTES_FROM_INT32(buf, this->u.intval);
		return sizeof(int);
		break;
	case fsubagentItem_STRING:
		{
			int len = strlen(this->u.stringval.buf);
			if (buflen <  len) {
				return -E2BIG;
			}
			memcpy(buf, this->u.stringval.buf, len);
			/* note: terminating NUL not copied to output buffer */
			return len;
		}
		break;
	default:
		;
	}

	return -EINVAL;
}


/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/


/**----------------------------------------------------------------------
 Create a fsubagentRow.
 @param constraints = stable pointer to row constraints
 @return initialized fsubagentRow * on success, NULL on failure
-----------------------------------------------------------------------*/
fsubagentRow *fsubagentRow_new(const fsubagentRowConstraints *constraints)
{
	int err;
	int i;
	enum fsubagentItemType kind;

	fsubagentRow *this = malloc(sizeof(*this));
	DPRINT(("fsubagentItem_new(%p): entry\n", constraints));
	if (!this)
		return NULL;
    err = pthread_mutex_init(&this->mutex, NULL);
    if (err) {
		free(this);
        return NULL;
	}
	this->constraints = constraints;
	this->nvars = constraints->maxoid;

	this->vars = malloc(this->nvars * sizeof(fsubagentItem *));
	if (!this->vars) {
		pthread_mutex_destroy(&this->mutex);
		free(this);
		return NULL;
	}

	for (i=0; (kind = constraints->format[i]) != 0; i++) {
		fsubagentItem *var = NULL;
		if (kind != fsubagentItem_UNDEF) {
			var = fsubagentItem_new(kind);
			if (!var)
				return NULL;
		}
		this->vars[i] = var;
	}

	DPRINT(("fsubagentItem_new(%p): success, returning %p\n", constraints, this));
	return this;
}

/**----------------------------------------------------------------------
 Delete a fsubagentRow and release any resources it has allocated.
-----------------------------------------------------------------------*/
void fsubagentRow_delete(fsubagentRow *this)
{
	DPRINT(("fsubagentItem_delete(%p): entry\n", this));
	if (this->vars) {
		int i;
		for (i=0; i<this->nvars; i++)
			if (this->vars[i])
				fsubagentItem_delete(this->vars[i], this->constraints->format[i]);
		free(this->vars);
		this->vars = NULL;
	}
	pthread_mutex_destroy(&this->mutex);
	DPRINT(("fsubagentItem_delete(%p): done\n", this));
	free(this);
}

/**----------------------------------------------------------------------
 Load new current row contents from buffer.
 Must obey constraints this row was created with.
 @param buf Buffer to load from.
 @param buflen Number of bytes of data in buffer.
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
int fsubagentRow_fromBuf(fsubagentRow *this, const char *buf, int buflen)
{
	int retval = 0;

	fsubagentRow_lock(this);

	/* format: 
	 * RowBody := ItemRecord+
	 * ItemRecord := magic(2) len(2) index(2) ItemBody
	 */

	while (buflen > 0) {
		int len, index, magic;
		int err;

		/* Read fixed-width part of item record */
		if (buflen < 6) {
			retval = EINVAL;
			break;
		}
		magic = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 2;
		if (magic != fsubagentRow_MAGIC)
			return EINVAL;

		len = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		index = BE_BYTES_TO_INT16v(buf);
		buf += 2;
		buflen -= 4;

		if (len < 1 || len > buflen) {
			retval = EINVAL;
			break;
		}

		if (index < 0 || index >= this->nvars) {
			retval = EINVAL;
			break;
		}

		/* Read in variable-length portion */
		err = fsubagentItem_fromBuf(this->vars[index], 
			this->constraints->format[index], buf, len);
		if (err) {
			retval = err;
			break;
		}
		buf += len;
		buflen -= len;
	}

	fsubagentRow_unlock(this);
	return retval;
}

/* See comment in fsubagentRow_fromBuf for format of row body */
#define ROW_HDR_LEN 6

/**----------------------------------------------------------------------
 Save new current row contents to buffer.
 @param buf Buffer to save to.
 @param bufsize Size of buffer in bytes.
 @param bufused Pointer to where to store number of buffer bytes used.
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
int fsubagentRow_toBuf(fsubagentRow *this, char *buf, int bufsize, int *bufused)
{
	/* See comment in fsubagentRow_fromBuf for format */
	int i;
	int retval = 0;

	*bufused = 0;

	fsubagentRow_lock(this);

	for (i=0; i<this->nvars; i++) {
		int len;
		/* skip to next occupied index.  Big gaps are uncommon. */
		if (!this->vars[i]) continue;

		/* Check to make sure we're not about to overflow the output buffer */
		if (*bufused + ROW_HDR_LEN >= bufsize) {
			retval = E2BIG;
			break;
		}

		/* Save the fixed-width portion */
		/* Leave the length field blank until we know what to put there */
		BE_BYTES_FROM_INT16(buf, fsubagentRow_MAGIC);
		BE_BYTES_FROM_INT16(buf+4, i);

		/* Save the variable-length portion */
		len = fsubagentItem_toBuf(this->vars[i], this->constraints->format[i], buf+ROW_HDR_LEN, bufsize-ROW_HDR_LEN - *bufused);
		if (len < 0) {
			retval = -len;
			break;
		}
		/* If there was nothing there, skip this record */
		if (len == 0) {
			continue;
		}
		/* Finally write the length field */
		BE_BYTES_FROM_INT16(buf+2, len);
		/*printf("fsubagentRow_toBuf: len 0x%x, index 0x%x\n", len, i);*/

		/* and skip to next record */
		buf += len + ROW_HDR_LEN;
		*bufused += len + ROW_HDR_LEN;
	}

	fsubagentRow_unlock(this);
	/*printf("fsubagentRow_toBuf: bufused %d\n", *bufused);*/
	return retval;
}

/**----------------------------------------------------------------------
 Set an integer variable in the row.
 @param index which variable in row to set (1 or higher)
 @param val value to set it to
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
int fsubagentRow_setInt(fsubagentRow *this, int index, int val)
{
	int retval = 0;

	DPRINT(("fsubagentItem_setInt(%p, %d, %d): entry\n", this, index, val));
	if (!this) {
		DPRINT(("fsubagentItem_setInt(%p, %d, %d): null pointer\n", this, index, val));
		assert(this != NULL);
		return EINVAL;
	}

	if (index < 1 
	|| index > this->constraints->maxoid 
	|| this->constraints->format[index-1] != fsubagentItem_INT) {
		DPRINT(("fsubagentItem_setInt(%p, %d, %d): bad index or type\n", this, index, val));
		return EINVAL;
	}

	fsubagentRow_lock(this);

	/* Set its value */
	this->vars[index-1]->u.intval = val;

	fsubagentRow_unlock(this);
	DPRINT(("fsubagentItem_setInt(%p, %d, %d): done\n", this, index, val));
	return retval;
}

/**----------------------------------------------------------------------
 Add the given signed increment to an integer variable in the row.
 @param index which variable in row to set (1 or higher)
 @param val value to add to it
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
int fsubagentRow_addInt(fsubagentRow *this, int index, int val)
{
	int retval = 0;

	DPRINT(("fsubagentItem_addInt(%p, %d, %d): entry\n", this, index, val));
	if (index < 1 
	|| index > this->constraints->maxoid 
	|| this->constraints->format[index-1] != fsubagentItem_INT)
		return EINVAL;

	fsubagentRow_lock(this);

	/* Add to its value */
	this->vars[index-1]->u.intval += val;

	fsubagentRow_unlock(this);
	DPRINT(("fsubagentItem_addInt(%p, %d, %d): done\n", this, index, val));
	return retval;
}

/**----------------------------------------------------------------------
 Set a string variable in the row.  Can only be done once.
 @param index which variable in row to set (1 or higher)
 @param val pointer to where to copy string value from
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
int fsubagentRow_setString(fsubagentRow *this, int index, const char *val)
{
	int retval = 0;

	DPRINT(("fsubagentItem_setString(%p, %d, %s): entry\n", this, index, val));
	if (index < 1 
	|| index > this->constraints->maxoid 
	|| this->constraints->format[index-1] != fsubagentItem_STRING)
		return EINVAL;

	fsubagentRow_lock(this);

	/* Set its value */
	retval = fsubagentItem_copyString(this->vars[index-1], val, strlen(val));

	fsubagentRow_unlock(this);
	DPRINT(("fsubagentItem_setString(%p, %d, %s): done\n", this, index, val));
	return retval;
}

/**----------------------------------------------------------------------
 Retrieve an integer variable in the row.
 @param index which variable in row to get (1 or higher)
 @param pval pointer to where to store value
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
int fsubagentRow_getInt(fsubagentRow *this, int index, int *pval)
{
	int retval = 0;

	index--;

	if (index < 0 
	|| index > this->constraints->maxoid) {
		/* FIXME: is this upper limit right? */
		DPRINT(("fsubagentItem_getInt(%p, %d, %p): failed: index out of range 0..%d\n", 
			this, index, pval, this->constraints->maxoid));
		return EINVAL;
	}
	if (this->constraints->format[index] != fsubagentItem_INT) {
		DPRINT(("fsubagentItem_getInt(%p, %d, %p): failed: bad type\n", this, index, pval));
		return EINVAL;
	}

	fsubagentRow_lock(this);

	*pval = this->vars[index]->u.intval;

	fsubagentRow_unlock(this);
	DPRINT(("fsubagentItem_getInt(%p, %d, %p): done, val %d\n", this, index, pval, *pval));
	return retval;
}

/**----------------------------------------------------------------------
 Retrieve a string variable in the row.
 @param index which variable in row to set (1 or higher)
 @param pp where to save pointer to const string inside row
 @return 0 on success, Unix error code on failure
-----------------------------------------------------------------------*/
int fsubagentRow_getString(fsubagentRow *this, int index, const char **p)
{
	int retval = 0;

	index--;

	if (index < 0 
	|| index > this->constraints->maxoid 
	|| this->constraints->format[index] != fsubagentItem_STRING)
		return EINVAL;

	fsubagentRow_lock(this);

	*p = this->vars[index]->u.stringval.buf;

	fsubagentRow_unlock(this);
	return retval;
}

/**----------------------------------------------------------------------
 Lock the row against access by other threads.
-----------------------------------------------------------------------*/
void fsubagentRow_lock(fsubagentRow *this)
{
	pthread_mutex_lock(&this->mutex);
}

/**----------------------------------------------------------------------
 Unlock the row.
-----------------------------------------------------------------------*/
void fsubagentRow_unlock(fsubagentRow *this)
{
	pthread_mutex_unlock(&this->mutex);
}
