/*--------------------------------------------------------------------------
 code to extend UCD-SNMP agent for DISAPPEARING-SNMP-MIB
 Compiled inside the ucd-snmp source tree 
--------------------------------------------------------------------------*/

/* I'd rather include the following two .h files within fsubagent_utils.h,
 * which needs types they define, but doing that seems to cause problems
 * (config.h doesn't protect itself against multiple inclusion, etc.).
 * Looks like any .c file that includes fsubagent_utils.h must
 * first include these two .h files.  Sigh.
 */
#include "config.h"
#include "mibincl.h"

#include "fsubagent_utils.h"

#include <limits.h>
#include <stdio.h>
#include <time.h>

#include "dprint.h"

/* How long to cache data read from disk. */
int fsubagent_utils_cacheSeconds = 10;

/* The object that grabs data from the file written by the monitored daemon */
fsubagent *fsubagent_utils_subagent;

#ifdef USE_DPRINT
dprint_oid(const char *label, oid *name, int len)
{
	int i;
	DPRINT(("%s: len %d, name ", label, len));
	for (i=0; i<(oid)len; i++)
		DPRINT(("%d ", name[i]));
	DPRINT(("\n"));
}
#else
#define dprint_oid(a, b, c)
#endif

/**
 * check currency of OID data and refresh if necessary
 * FIXME: uses a static variable
 */
void fsubagent_utils_update(int force) 
{
	int err;

	/* Don't check more than once every fsubagent_utils_cache_seconds+1 seconds */
	static time_t lastUpdate = -INT_MAX;
	time_t now = time(0);
	DPRINT(("fsubagent_utils_update(%d): now %d lastUpdate %d\n", force, now, lastUpdate));
	if ((lastUpdate != -INT_MAX) && !force && (now < lastUpdate + fsubagent_utils_cacheSeconds))
		return;

	err = fsubagent_load(fsubagent_utils_subagent);
	if (err) {
		DPRINT(("fsubagent_utils_update(%d): error %d loading.\n", force, err));
	} else {
		DPRINT(("fsubagent_utils_update(%d): loaded ok.\n", force));
		lastUpdate = now;
	}
}

/* Handle a read request of the given variable. 
 * FIXME: uses a static variable
 */
static unsigned char *
fsubagent_utils_getFromRow(int index, fsubagentRow *row, struct variable *vp, size_t *var_len)
{
	static int ival;
	char *sval;
	int err;

	DPRINT(("fsubagent_utils_getFromRow(%d, %p, %p, %p): vp->type %d\n", index, row, vp, var_len, vp->type));

	switch (vp->type) {
	case	ASN_COUNTER:
	case	ASN_GAUGE:
	case	ASN_INTEGER:
	case	ASN_TIMETICKS:
		ival = 0;
		err = fsubagentRow_getInt(row, index, &ival);
		if (err) {
			DPRINT(("fsubagent_utils_getFromRow(%p, %p, %p): int fail, err %d\n", row, vp, var_len, err));
			return NULL;
		}
		DPRINT(("fsubagent_utils_getFromRow(%p, %p, %p): int done, val %d\n", row, vp, var_len, ival));
		*var_len = sizeof(int);
		return (unsigned char *)&ival;

	case	ASN_OCTET_STR:
		sval = NULL;
	
		fsubagentRow_getString(row, index, (const char **)&sval);
		/* Any strings in row are null-terminated ascii */
		if (sval)
			*var_len = strlen(sval);
		DPRINT(("fsubagent_utils_getFromRow(%p, %p, %p): string done, val %s\n", row, vp, var_len, 
			(sval ? sval : "null")));
		return (unsigned char *)sval;

	default:
		break;
	}
	*var_len = 0;
	DPRINT(("fsubagent_utils_getFromRow(%p, %p, %p): failed\n", row, vp, var_len));
	return NULL;
}

/* Called by wrapper functions.
 * Check OID against given row, then get variable if appropriate. 
 */
unsigned char *
fsubagent_utils_getFromScalarRow(
	fsubagentRow		*row,
	struct variable *vp, 
	oid				*name, 
	size_t			*length, 
	int				exact, 
	size_t			*var_len, 
	WriteMethod	**write_method)
{
	int colindex;
	DPRINT(("getFromScalarRow(%p,%p,%d,%d,): top\n", row, vp, *length, exact));

	/* Get latest data, if needed. */
	fsubagent_utils_update(FALSE);
	if (header_generic(vp,name,length,exact,var_len,write_method) == MATCH_FAILED) {
		dprint_oid("getFromScalarRow match fail", name, *length);
		return NULL;
	}
	colindex = name[*length - 2];
	DPRINT(("getFromScalarRow: name[%d] = colindex = %d\n", *length, colindex));
	return fsubagent_utils_getFromRow(colindex, row, vp, var_len);
}

/* Called by wrapper functions.
 * Check OID against given 1-index table, then get variable if appropriate. 
 */
unsigned char *
fsubagent_utils_getFromTable1(
	fsubagentArray	*arr,
	struct variable *vp, 
	oid				*name, 
	size_t			*length, 
	int				exact, 
	size_t			*var_len, 
	WriteMethod	**write_method)
{
	int colindex, rowindex;
	int maxrow;
	fsubagentRow *row;

	DPRINT(("getFromTable1(%p,%p,%d,%d,): top\n", arr, vp, *length, exact));

	/* Refresh our copy of the data so limit checks are up to date. */
	fsubagent_utils_update(FALSE);

	maxrow = fsubagentArray_getMaxRow(arr);
	if (header_simple_table(vp, name, length, exact, var_len, write_method, maxrow) == MATCH_FAILED)
		return NULL;
	colindex = name[*length - 2];
	rowindex = name[*length - 1];
	row = fsubagentArray_getRow(arr, rowindex);
	return fsubagent_utils_getFromRow(colindex, row, vp, var_len);
}

/* Called by wrapper functions.
 * Check OID against given 2-index table, then get variable if appropriate. 
 * More complicated because there's no 'header_simple_table2' routine.
 * See http://www.csc.liv.ac.uk/~daves/Misc/UCD/07-table2.html
 * and compare with header_simple_table() in util_funcs.c
 * and header_hrpartition() in ucd-snmp-4.2/agent/mibgroup/host/hr_partition.c
 * FIXME: this needs cleaning up!
 */
unsigned char *
fsubagent_utils_getFromTable2(
	fsubagentArray2	*arr2,
	struct variable *vp, 
	oid				*name, 
	size_t			*length, 
	int				exact, 
	size_t			*var_len, 
	WriteMethod	**write_method)
{
	fsubagentArray *arr;
	fsubagentRow *row;
	int i;
	int rtest; /* -1  if name < vp->name, 1 if name > vp->name, 0 otherwise.  */
	oid newname[MAX_OID_LEN];
	const int nindices = 2;		/* number of OID components in this table's indices */
	int colindex, rowindex1, rowindex2;

	DPRINT(("getFromTable2(%p,%p,%d,%d,): top\n", arr2, vp, *length, exact));

	/* Refresh our copy of the data so limit checks are up to date. */
	fsubagent_utils_update(FALSE);

	memmove(newname, name, (int) (*length) * sizeof(oid));

	if (!exact && (int)(*length) < (int) vp->namelen + nindices) {
		DPRINT(("getFromTable2: !exact, request is missing indices; supplying the missing indices.\n"));
		while ((int)*length < (int) vp->namelen + nindices)
			newname[(*length)++] = 0;
	}

	/* Compare the requested OID with the OID of the top of this table */
	/* FIXME: isn't there some library function for this? */
	rtest = 0;
	for (i = 0; i < (int) vp->namelen && i < (int) (*length); i++) {
		if (newname[i] != vp->name[i]) {
			rtest = (newname[i] < vp->name[i]) ? -1 : 1;
			DPRINT(("getFromTable2: newname[%d] = %d, vp->name[%d] = %d\n", i, newname[i], i, vp->name[i]));
			break;
		}
	}
	/* Reject fetch request if that part of OID doesn't match well enough */
	if (rtest > 0) {
		/* is this needed?!  Should never happen. */
		*var_len = 0;
		DPRINT(("getFromTable2(exact %d): fail: name beyond our part of OID space\n", exact));
		return NULL;
	}
	if (rtest == 0 && (int) *length > (int) (vp->namelen + nindices)) {
		DPRINT(("getFromTable2(exact %d): fail: name inside our part of OID space, but too long\n", exact));
		return NULL;
	}
	if (exact
	&& (rtest || (int) *length != (int) (vp->namelen + nindices))) {
		*var_len = 0;
		/* Reject if exact match desired and name doesn't match exactly or is wrong length */
		DPRINT(("getFromTable2(exact %d): fail: rtest %d, len %d, want len %d\n", 
			exact, rtest, *length, vp->namelen + nindices));
		return NULL;
	}
	if (!exact 
	&& (rtest == -1)) {
		DPRINT(("getFromTable2: Name lower than our area; setting to first OID in our area\n"));
		memmove(newname, vp->name, (int) (vp->namelen) * sizeof(oid));
		for (i = 0; i < nindices; i++)
			newname[vp->namelen + i] = 0;
		*length = vp->namelen + nindices;
	}

	/* Now make sure that the resulting OID corresponds
	 * to a real data object.  This code is specialized for
	 * a two-index table.
	 */
	rowindex1 = newname[*length - 2];
	rowindex2 = newname[*length - 1];

	row = NULL;
	/* advance rowindex1 to first valid value */
	do {
		arr = fsubagentArray2_getArray(arr2, rowindex1);
		if (!arr && !exact && rowindex1 < fsubagentArray2_getMaxArray(arr2))
			rowindex1++;
		else
			break;
	} while (1);
	if (!arr) {
		DPRINT(("getFromTable2(exact %d): first index %d out of range, failure\n", exact, rowindex1));
		return NULL;
	}
	do {
		if (!exact) {
			/* It's a GETNEXT request; must increment the OID to the next valid one. */
			rowindex2++;
			if (rowindex2 > fsubagentArray_getMaxRow(arr)) {
				/* If we've hit the max row, wrap to next array. */
				rowindex1++;
				rowindex2 = 1;
				arr = fsubagentArray2_getArray(arr2, rowindex1);
				if (!arr) {
					DPRINT(("getFromTable2(exact %d): incremented first index %d out of range\n", exact, rowindex1));
					break;
				}
			}
		}
		row = fsubagentArray_getRow(arr, rowindex2);
	} while (!row && !exact);

	newname[*length - 2] = rowindex1;
	newname[*length - 1] = rowindex2;

	if (!row) {
		if (!exact)
			memmove(name, newname, (*length) * sizeof(oid));
		if (var_len)
			*var_len = 0;
		DPRINT(("getFromTable2(exact %d): failure\n"));
		return NULL;
	}
	DPRINT(("getFromTable2(exact %d): found object; indices %d %d\n", exact, rowindex1, rowindex2));

	/* We located the data object named by the OID.  Save its name. */
	memmove(name, newname, (*length) * sizeof(oid));
	if (write_method)
		*write_method = 0;

	/* Retrieve its value. */
	colindex = name[*length - 3];
	return fsubagent_utils_getFromRow(colindex, row, vp, var_len);
}

