#!/bin/perl
# Usage: perl mib2h.pl your-mib-text-file > your-parsed-mib.h
# Script to parse a text MIB file and produce a .h file
# suitable for use with either C or C++ which has the
# numerical constants for the SNMP OID's in the MIB file.
# A bit fragile, but since you're the one writing the input file, that's ok.
# Written for use with fsubagent by Dan Kegel for Omniva Policy Systems

# Outputs one #define for each object giving its OID value
# relative to its parent, plus a comment describing its type.

foreach (<>) {
	if (/(\S+)\s+DEFINITIONS/) {
		$prefix = $1;
		# dashes are not allowed in C; use underscores instead
		$prefix =~ s/-/_/g;
		# shorten identifier -- if it ends in _MIB, strip off that suffix
		$prefix =~ s/_MIB$//;
	} elsif (/(\S+)\s+OBJECT-TYPE/) {
		$id = $1;
		$in_object_type = 1;
		$in_textual_convention = 0;
	} elsif (/(\S+)\s+OBJECT IDENTIFIER\s*::=\s+{\s+(\S+)\s+(\d+)\s*}/) {
		$id = $1;
		$parent = $2;
		$val = $3;
		$type = "";
		push(@vars, "$id:$parent:$val:$type");
		$id = ""; $parent = ""; $val = ""; $type = "";
	} elsif (/^(\S+)\s+::=\s+TEXTUAL-CONVENTION/) {
		$id = $1;
		$in_textual_convention = 1;
	} elsif (/^(\S+)\s+::=\s+INTEGER \(.*\)/) {
		# Handle constrained integers as if they were textual conventions
		$id = $1;
		$type = "Integer";
		$conventions{$id} = $type;
	} elsif (/^(\S+)\s+::=\s+SEQUENCE/) {
		$id = $1;
		$isSequence{$id} = 1;
		$in_object_type = 0;
	} elsif (/^[^-\s]/) {
		# lines with a nonblank noncomment in column one are the start of something new
		$in_object_type = 0;
		$in_textual_convention = 0;
	}

	if ($in_object_type) {
		if (/SYNTAX (\S+)/) {
			$type = $1;
		}
		if (/::=\s+{\s+(\S+)\s+(\d+)\s*}/) {
			$parent = $1;
			$val = $2;
			push(@vars, "$id:$parent:$val:$type");
			$id = ""; $parent = ""; $val = ""; $type = "";
			$in_object_type = 0;
		}
	}
	if ($in_textual_convention) {
		if (/SYNTAX\s+(\S+)/) {
			$type = $1;
			$conventions{$id} = $type;
			$in_textual_convention = 0;
		}
	}
}

# Figure out max identifier length, note user defined types
$maxlen = 0;
foreach (@vars) {
	($id, $parent, $val, $type) = split(/:/);
	if (length($id) > $maxlen) {
		$maxlen = length($id);
	}
}

print "
/* Generated by mib2h.pl */
#ifndef ".$prefix."_INCLUDED
#define ".$prefix."_INCLUDED

";

foreach (@vars) {
	($id, $parent, $val, $type) = split(/:/);
	printf("#define ".$prefix."_%-".$maxlen."s %5d /* $type", $id, $val);
	if ($conventions{$type} ne "") {
		print " aka " . $conventions{$type};
	}
	print ", in $parent */\n";
}
print "\n";

# Figure out format string for every node that has non-oid children
# First, collect variables by parent
foreach (@vars) {
	($id, $parent, $val, $type) = split(/:/);
	$tabletype = $type;
	substr($tabletype,0,1) =~ tr/a-z/A-Z/;
	# note: $type eq SEQUENCE means it's a SEQUENCE OF something;
	# $isSequence{$type} means it's a SEQUENCE.  Bleh.
	# Want to exclude both, since they're not leaf nodes.
	if ($type ne "" && $type ne "SEQUENCE" && !$isSequence{$tabletype}) {
		if ($conventions{$type} ne "") {
			$type = $conventions{$type};
		}
		if ($children{$parent} ne "") {
			$children{$parent} .= ",";
		}
		$children{$parent} .= "$val=$type=$id";
	}
}

# Print out leaf node types.
foreach $parent (sort(keys(%children))) {
	local(%types, %ids);
	#print "/* $parent : ".$children{$parent}." */\n";
	foreach (split(/,/, $children{$parent})) {
		local($val, $type, $id) = split(/=/);
		$types{$val} = $type;
		$ids{$val} = $id;
	}
	# Format string has one char for each index starting with 1
	# Char is . for empty, s for string, and i for integer
	$index = 1;
	$format = "";
	foreach (sort { $a - $b } keys(%types)) {
		# Represent any gaps in index space
		while ($_ > $index) {
			$format .= ".";
			$index++;
		}
		# Represent the variable at this index
		$type = $types{$_};
		$type =~ tr/A-Z/a-z/;
		if ($type eq "integer") {
			$format .= "i";
		} elsif ($type eq "counter") {
			$format .= "i";
		} elsif ($type eq "gauge") {
			$format .= "i";
		} elsif ($type eq "timeticks") {
			$format .= "i";
		} elsif ($type eq "displaystring") {
			$format .= "s";
		} else {
			die "Unrecognized type $type";
		}
		$index++;
	}
	print "#define ".$prefix."_".$parent."_FORMAT \"$format\"\n";

	# Print initializers for the static 'struct variable2' array.
	print "#define ".$prefix."_VARS_".$parent." \\\n";
	foreach (sort { $a - $b } keys(%types)) {
		my($id) = $ids{$_};
		my($type) = $types{$_};
		$type =~ tr/A-Z/a-z/;
		if ($type eq "integer") {
			$c_type = "ASN_INTEGER";
		} elsif ($type eq "counter") {
			$c_type = "ASN_COUNTER";
		} elsif ($type eq "gauge") {
			$c_type = "ASN_GAUGE";
		} elsif ($type eq "timeticks") {
			$c_type = "ASN_TIMETICKS";
		} elsif ($type eq "displaystring") {
			$c_type = "ASN_OCTET_STR";
		} else {
			die "Unrecognized type $type";
		}
		print "{ 0, $c_type, RONLY, handle_$parent, 1, {".$prefix."_".$id."} },\\\n";
	}
	print "\n";

	# Generate C++ wrapper class for this row.
	# struct DIMIB_fooWrapper 
	# {
	# private:
	#  fsubagentRow *m_row;
	# public:
	#  DIMIB_fooWrapper(fsubagentRow *row) { m_row = row; }
	#  int set_mumble(int val) { return fsubagentRow_setInt(m_row, DIMIB_mumble, val); }
	#  int add_mumble(int val) { return fsubagentRow_addInt(m_row, DIMIB_mumble, val); }
	#  int get_mumble(int *pval) { return fsubagentRow_setInt(m_row, DIMIB_mumble, pval); }
	#  int set_foo(const char *cstr) { return fsubagentRow_setString(m_row, DIMIB_foo, cstr); }
	#  int get_foo(const char **pcstr) { return fsubagentRow_getString(m_row,DIMIB_foo,pcstr);}
	# }
	print "
struct ".$prefix."_".$parent."Wrapper 
#ifdef __cplusplus
{
private:
	fsubagentRow *m_row;
public:
	".$prefix."_".$parent."Wrapper(fsubagentRow *row) { m_row = row; }
	fsubagentRow *getfsubagentRow() { return m_row; }
	/* Accessors return 0 on success, Unix error code on failure. */
";
	foreach (sort { $a - $b } keys(%types)) {
		# Generate setter and getter for the variable at this index
		$id = $ids{$_};
		$type = $types{$_};
		$type =~ tr/A-Z/a-z/;
		if ($type eq "integer" || $type eq "counter" || $type eq "gauge" || $type eq "timeticks") {
			print 
"	int set_$id(int val) { return fsubagentRow_setInt(m_row, ".$prefix."_$id, val); }
	int add_$id(int val) { return fsubagentRow_addInt(m_row, ".$prefix."_$id, val); }
	int get_$id(int *pval) { return fsubagentRow_getInt(m_row, ".$prefix."_$id, pval); }
";
		} elsif ($type eq "displaystring") {
			print 
"	int set_$id(const char *cstr) { return fsubagentRow_setString(m_row, ".$prefix."_$id, cstr); }
	int get_".$id."Ptr(const char **pcstr) { return fsubagentRow_getString(m_row, ".$prefix."_$id, pcstr); }
";
		}
	}
	print "
}
#else
; typedef struct ".$prefix."_".$parent."Wrapper ".$prefix."_".$parent."Wrapper 
#endif
;

";

}

printf "
#endif
";

# Print out macro to be used in mib implementation
# (currently disappearing.c) to invoke the init routine
# for each table handler.  Even though this script doesn't
# currently generate table handlers like it should,
# it does generate calls to the table init routines,
# which will cause a build failure if you forget to
# hand-code a table handler, so at least you are reminded
# automatically that you have to do some hand coding...
print "#define $prefix"."_INIT \\\n";
foreach $parent (sort(keys(%children))) {
	push(@inits, "\tinit_$parent()");
}
print join("; \\\n", @inits)."\n";
