#ifndef portable_H
#define portable_H

/* Byte-order portability macros/functions.  Used by fsubagent
 * to control the byte order of the files it writes, so files
 * can be exchanged between big-endian and little-endian machines,
 * and so Java code could be written to read or write these files.
 *
 * Ironically, this code itself isn't portable C:
 * 1. it uses the 'inline' keyword, which isn't in ANSI C.  If this causes 
 *    an error, just delete that word, it should compile and work fine.
 * 2. it expects 'int x = (c << 24)' to widen char c before shifting,
 *    which some compilers might not do
 * 3. most callers of these macros call with signed char *
 *    arguments, and a pedantic compiler will catch the mismatch.
 * Not sure it's worth fixing these issues until they actually bite
 * somebody.
 */

typedef unsigned int mem_uint32_t;
typedef unsigned short mem_uint16_t;

/**
 Given a 32 bit integer, calculate what the nth byte of its big-endian
 representation would be.<br>
 Works regardless of whether local computer is little-endian or big-endian.<br>
 Example: 
 to convert 'myint32' to a big-endian byte stream, you could use<br>
   for (i=0;i!=sizeof(myint32);i++) obuf[i] = BE_BYTE_FROM_INT32(myint32, i);
 */
#define BE_BYTE_FROM_INT32(val,n) (((unsigned)(val) >> ((3 - n) * 8)) & 0xff) 

/** Buffer version of BE_BYTE_FROM_INT32 */
static inline void BE_BYTES_FROM_INT32(unsigned char *obuf, mem_uint32_t myint32) {
	size_t i;
	for (i=0;i!=sizeof(myint32);i++) obuf[i] = BE_BYTE_FROM_INT32(myint32, i);
}

/**
 Given a 16 bit integer, calculate what the nth byte of its big-endian
 representation would be.<br>
 Works regardless of whether local computer is little-endian or big-endian.<br>
 Example: 
 to convert 'myint16' to a big-endian byte stream, you could use<br>
   for (i=0;i!=sizeof(myint16);i++) obuf[i] = BE_BYTE_FROM_INT16(myint16, i);
 */
#define BE_BYTE_FROM_INT16(val,n) (((unsigned)(val) >> ((1 - n) * 8)) & 0xff) 

/** Buffer version of BE_BYTE_FROM_INT16 */
static inline void BE_BYTES_FROM_INT16(unsigned char *obuf, mem_uint16_t myint16) {
	size_t i;
	for (i=0;i!=sizeof(myint16);i++) obuf[i] = BE_BYTE_FROM_INT16(myint16, i);
}

/***********************************************************************/

/**
 Given a big-endian byte stream, reconstruct a 16 bit integer.
 Works regardless of whether local computer is little-endian or big-endian.<br>
 Example: 
 to convert a big-endian byte stream back to 'myint16', you could use<br>
   myint16 = BE_BYTES_TO_INT16(ibuf[0], ibuf[1]);
 */
static inline mem_uint16_t BE_BYTES_TO_INT16(unsigned char byte0, unsigned char byte1)
{
	return ((byte0 << 8) | (byte1));
}

/** buffer version of BE_BYTES_TO_INT16 */
static inline mem_uint16_t BE_BYTES_TO_INT16v(const unsigned char* bytes)
{
	return BE_BYTES_TO_INT16(bytes[0], bytes[1]);
}

/**
 Given a big-endian byte stream, reconstruct a 32 bit integer.
 Works regardless of whether local computer is little-endian or big-endian.<br>
 Example: 
 to convert a big-endian byte stream back to 'myint32', you could use<br>
   myint32 = BE_BYTES_TO_INT32(ibuf[0], ibuf[1], ibuf[2], ibuf[3]);
 */
static inline mem_uint32_t BE_BYTES_TO_INT32(unsigned char byte0, unsigned char byte1,
									  unsigned char byte2, unsigned char byte3)
{
	return ((byte0 << 24) | (byte1 << 16) | (byte2 << 8) | (byte3));
}

/** buffer version of BE_BYTES_TO_INT32 */
static inline mem_uint32_t BE_BYTES_TO_INT32v(const unsigned char *bytes)
{
	return BE_BYTES_TO_INT32(bytes[0], bytes[1], bytes[2], bytes[3]);
}

#endif
